// GeekOS interrupt handling data structures and functions
// Copyright (c) 2001, David H. Hovemeyer <daveho@cs.umd.edu>
// $Revision: 1.1.1.1 $

// This is free software.  You are permitted to use,
// redistribute, and modify it as specified in the file "COPYING".

#include "idt.h"	// x86-specific int handling stuff
#include "screen.h"
#include "kassert.h"
#include "int.h"

// Defined in lowlevel.asm.
unsigned long Get_Current_EFLAGS( void );

// ----------------------------------------------------------------------
// Private functions and data
// ----------------------------------------------------------------------

//
// A dummy interrupt handler function.
// Ensures that the low-level interrupt code always
// has a handler to call.
//
static void Dummy_Interrupt_Handler( struct Interrupt_State* state )
{
    Print( "*** Unexpected interrupt! ***\n" );
    Dump_Interrupt_State( state );
    STOP();
}

static void Print_Selector( const char* regName, unsigned int value )
{
    Print( "%s: index=%d, ti=%d, rpl=%d\n",
	regName, value >> 3, (value >> 2) & 1, value & 3 );
}

// ----------------------------------------------------------------------
// Public functions
// ----------------------------------------------------------------------

//
// Initialize the interrupt system.
//
void Init_Interrupts( void )
{
    int i;

    // Low-level initialization.  Build and initialize the IDT.
    Init_IDT();

    // Initialize all entries of the handler table with a dummy handler.
    // This will ensure that we always have a handler function to call.
    for ( i = 0; i < NUM_IDT_ENTRIES; ++i ) {
	Install_Interrupt_Handler( i, Dummy_Interrupt_Handler );
    }

    // Re-enable interrupts
    Enable_Interrupts();
}

// Query whether or not interrupts are currently enabled.
Boolean Interrupts_Enabled( void )
{
    unsigned long eflags = Get_Current_EFLAGS();
    return (eflags & EFLAGS_IF) != 0;
}

// Dump interrupt state struct to screen
void Dump_Interrupt_State( struct Interrupt_State* state )
{
    unsigned int errorCode = state->errorCode;

    Print( "eax=%x ebx=%x ecx=%x edx=%x\n"
	   "esi=%x edi=%x ebp=%x\n"
	   "eip=%x cs=%x eflags=%x\n"
	   "Interrupt number=%d, error code=%d\n"
	   "index=%d, TI=%d, IDT=%d, EXT=%d\n",
	state->eax, state->ebx, state->ecx, state->edx,
	state->esi, state->edi, state->ebp,
	state->eip, state->cs, state->eflags,
	state->intNum, errorCode,
	errorCode >> 3, (errorCode >> 2) & 1, (errorCode >> 1) & 1, errorCode & 1
	);
    Print_Selector( "cs", state->cs );
    Print_Selector( "ds", state->ds );
    Print_Selector( "es", state->es );
    Print_Selector( "fs", state->fs );
    Print_Selector( "gs", state->gs );
}
