// GeekOS system calls
// Copyright (c) 2001, David H. Hovemeyer <daveho@cs.umd.edu>
// $Revision: 1.2 $

// This is free software.  You are permitted to use,
// redistribute, and modify it as specified in the file "COPYING".

#include "int.h"
#include "malloc.h"
#include "screen.h"
#include "user.h"
#include "string.h"
#include "kthread.h"
#include "keyboard.h"
#include "synch.h"
#include "syscall.h"

// ----------------------------------------------------------------------
// System call implementation functions
// ----------------------------------------------------------------------

static struct Mutex s_screenLock = MUTEX_INITIALIZER;
static struct Mutex s_keyboardLock = MUTEX_INITIALIZER;

// The null system call; does absolutely nothing
// except return a value
static int Sys_Null( struct Interrupt_State* state )
{
    return 0;
}

// Exit system call; kills the thread.
static int Sys_Exit( struct Interrupt_State* state )
{
    struct Kernel_Thread* current = g_currentThread;

    Kill_Thread( current );

    Disable_Interrupts();

    Detach_Thread( g_currentThread ); // send the thread to the reaper
    g_killCurrentThread = TRUE;     // don't make the thread runnable again
    g_needReschedule = TRUE;	    // pick a new thread

    return -1;
}

// Print a string to the screen using current attribute.
// Registers:
//   ebx - address of the string
//   ecx - the length of the string
static int Sys_Printstring( struct Interrupt_State* state )
{
    const void* userPtr = (const void*) state->ebx;
    unsigned int length = state->ecx;
    unsigned char* buf;

    // Make sure buf is a reasonable size.
    if ( length > 1024 )
	return -1;

    buf = Malloc_Atomic( length + 1 );
    if ( buf == 0 )
	return -1;

    if ( !Copy_From_User( buf, userPtr, length ) )
	return -1;
    buf[ length ] = '\0';

    Mutex_Lock( &s_screenLock );
    Print( "%s", buf );
    Mutex_Unlock( &s_screenLock );

    Free_Atomic( buf );

    KASSERT( Interrupts_Enabled() );

    return 0;
}

// Get a keycode from the keyboard and return it to user space.
static int Sys_Getkey( struct Interrupt_State* state )
{
    Keycode result;

    Mutex_Lock( &s_keyboardLock );

    // Note that this function can Wait()!
    result = Wait_For_Key_Atomic();

    Mutex_Unlock( &s_keyboardLock );

    return result;
}

// Set current video attribute.
// Ebx register contains the attribute to set.
static int Sys_Setattr( struct Interrupt_State* state )
{
    Mutex_Lock( &s_screenLock );
    Set_Current_Attr( state->ebx );
    Mutex_Unlock( &s_screenLock );

    return 0;
}

static int Sys_Fork( struct Interrupt_State* state )
{
	Print( "System Call : fork is called\n" );

	return 0;
}

static int Sys_Execve( struct Interrupt_State* state )
{
	Print( "System Call : execve is called\n" );

	return 0;
}

static int Sys_Open( struct Interrupt_State* state )
{
	Print( "System Call : open is called\n" );

	return 0;
}

static int Sys_Read( struct Interrupt_State* state )
{
	Print( "System Call : read is called\n" );

	return 0;
}

static int Sys_Write( struct Interrupt_State* state )
{
	Print( "System Call : write is called\n" );

	return 0;
}

static int Sys_Close( struct Interrupt_State* state )
{
	Print( "System Call : close is called\n" );

	return 0;
}

static int Sys_WaitPid( struct Interrupt_State* state )
{
	Print( "System Call : waitpid is called\n" );

	return 0;
}

static int Sys_GetPid( struct Interrupt_State* state )
{
	Print( "System Call : getpid is called\n" );

	return 0;
}

static int Sys_Create( struct Interrupt_State* state )
{
	Print( "System Call : create is called\n" );

	return 0;
}

static int Sys_Link( struct Interrupt_State* state )
{
	Print( "System Call : link is called\n" );

	return 0;
}

static int Sys_Unlink( struct Interrupt_State* state )
{
	Print( "System Call : unlink is called\n" );

	return 0;
}



// ----------------------------------------------------------------------
// Exported table of system call functions
// ----------------------------------------------------------------------

const Syscall g_syscallTable[] =
{
    &Sys_Null,
    &Sys_Exit,
    &Sys_Printstring,
    &Sys_Getkey,
    &Sys_Setattr,
	&Sys_Fork,
	&Sys_Execve,
	&Sys_Open,
	&Sys_Read,
	&Sys_Write,
	&Sys_Close,
	&Sys_WaitPid,
	&Sys_GetPid,
	&Sys_Create,
	&Sys_Link,
	&Sys_Unlink
    // add more syscall handlers here...
};

const unsigned int g_numSyscalls = sizeof(g_syscallTable) / sizeof(Syscall);
