// A library of useful functions for user-mode programs.

#include "libuser.h"

// Entry point.  Calls user program's Main() routine,
// then exits.  TODO: pass arguments to Main().
void _Entry( void )
{
    extern void Main( void );

    Main();
    exit();
}

// ----------------------------------------------------------------------
// Wrappers for system calls
// ----------------------------------------------------------------------

#define SYSCALL "int $0x90"

// system call numbers
#define SYS_NULL 0
#define SYS_EXIT 1
#define SYS_PRINTSTRING 2
#define SYS_GETKEY 3
#define SYS_SETATTR 4
#define SYS_FORK 5
#define SYS_EXECVE 6
#define SYS_OPEN 7
#define SYS_READ 8
#define SYS_WRITE 9
#define SYS_CLOSE 10
#define SYS_WAITPID 11
#define SYS_GETPID 12
#define SYS_CREATE 13
#define SYS_LINK 14
#define SYS_UNLINK 15

int Null( void )
{
    int num = SYS_NULL, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;
}

void exit( void )
{
    int num = SYS_EXIT;

    __asm__ __volatile__ (
	SYSCALL
	:
	: "a" (num)
    );
}

int Print_String( const char* message )
{
    int num = SYS_PRINTSTRING, rc;
    size_t len = strlen( message );

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)
	: "a" (num), "b" (message), "c" (len)
    );

    return rc;
}

Keycode Get_Key( void )
{
    int num = SYS_GETKEY, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)
	: "a" (num)
    );

    return (Keycode) rc;
}

void Set_Attr( unsigned int attrib )
{
    int num = SYS_SETATTR;

    __asm__ __volatile__ (
	SYSCALL
	:
	: "a" (num), "b" (attrib)
    );
}

int fork( void )
{
 	int num = SYS_FORK, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int execve( void )
{
 	int num = SYS_EXECVE, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int open( void )
{
 	int num = SYS_OPEN, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int read( void )
{
 	int num = SYS_READ, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int write( void )
{
 	int num = SYS_WRITE, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int close( void )
{
 	int num = SYS_CLOSE, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int waitpid( void )
{
 	int num = SYS_WAITPID, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int getpid( void )
{
 	int num = SYS_GETPID, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int create( void )
{
 	int num = SYS_CREATE, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int link( void )
{
 	int num = SYS_LINK, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

int unlink( void )
{
 	int num = SYS_UNLINK, rc;

    __asm__ __volatile__ (
	SYSCALL
	: "=a" (rc)	// system call returns value in eax
	: "a" (num)	// system call number specified in eax
    );

    return rc;

}

// ----------------------------------------------------------------------
// Useful functions
// ----------------------------------------------------------------------

static Boolean s_echo = TRUE;

void Echo( Boolean enable )
{
    s_echo = enable;
}

void Put_Char( int c )
{
    char buf[2];
    buf[0] = c;
    buf[1] = '\0';
    Print_String( buf );
}

void Read_Line( char* buf, size_t bufSize )
{
    size_t n = 0;
    Keycode k;
    Boolean done = FALSE;

    bufSize--;
    do {
    again:
	k = Get_Key();
	if ( (k & KEY_SPECIAL_FLAG) || (k & KEY_RELEASE_FLAG) )
	    goto again;

	k &= 0xff;
	if ( k == '\r' )
	    k = '\n';

	if ( s_echo )
	    Put_Char( k );

	if ( k == '\n' )
	    done = TRUE;

	if ( n < bufSize ) {
	    *buf++ = k;
	    ++n;
	}
    }
    while ( !done );

    *buf = '\0';
}
