// User mode context
// Copyright (c) 2001, David H. Hovemeyer <daveho@cs.umd.edu>
// $Revision: 1.1.1.1 $

// This is free software.  You are permitted to use,
// redistribute, and modify it as specified in the file "COPYING".

#include "gdt.h"
#include "malloc.h"
#include "mem.h"
#include "defs.h"
#include "int.h"
#include "kthread.h"
#include "tss.h"
#include "string.h"
#include "kassert.h"
#include "user.h"

// ------------------------------------------------------------
// Private data and functions
// ------------------------------------------------------------

// Most recently activated user context.
static struct User_Context* s_currentUserContext = 0;

// ------------------------------------------------------------
// Public functions
// ------------------------------------------------------------

// Create a new user context of given size
struct User_Context* Create_User_Context( unsigned long size )
{
    struct User_Context* context;
    int index;

    // Size must be a multiple of the page size
    size = Round_Up_To_Page( size );

    // Allocate memory for the user context
    Disable_Interrupts();
    context = (struct User_Context*) Malloc( sizeof(*context) );
    if ( context != 0 )
	context->memory = Malloc( size );
    Enable_Interrupts();

    if ( context == 0 || context->memory == 0 )
	goto fail;

    context->size = size;

    // Allocate an LDT descriptor for the user context
    context->ldtDescriptor = Allocate_Segment_Descriptor();
    if ( context->ldtDescriptor == 0 )
	goto fail;
    Init_LDT_Descriptor( context->ldtDescriptor, context->ldt, NUM_USER_LDT_ENTRIES );
    index = Get_Descriptor_Index(context->ldtDescriptor);
    context->ldtSelector = Selector( KERNEL_PRIVILEGE, TRUE, index );

    // Initialize code and data segments within the LDT
    Init_Code_Segment_Descriptor(
	&context->ldt[0],
	(unsigned long) context->memory,
	size / PAGE_SIZE,
	USER_PRIVILEGE
    );
    Init_Data_Segment_Descriptor(
	&context->ldt[1],
	(unsigned long) context->memory,
	size / PAGE_SIZE,
	USER_PRIVILEGE
    );
    context->csSelector = Selector( USER_PRIVILEGE, FALSE, 0 );
    context->dsSelector = Selector( USER_PRIVILEGE, FALSE, 1 );

    // Nobody is using this user context yet
    context->refCount = 0;

    // Success!
    return context;

fail:
    // We failed; release any allocated memory
    Disable_Interrupts();
    if ( context != 0 ) {
	if ( context->memory != 0 )
	    Free( context->memory );
	Free( context );
    }
    Enable_Interrupts();

    return 0;
}

// Destroy the given user context.
void Destroy_User_Context( struct User_Context* userContext )
{
    KASSERT( userContext->refCount == 0 );

    // Free the context's LDT descriptor
    Free_Segment_Descriptor( userContext->ldtDescriptor );

    // Free the context's memory
    Disable_Interrupts();
    Free( userContext->memory );
    Free( userContext );
    Enable_Interrupts();
}

// Attach a user context to the given thread.
// Called with interrupts enabled.
void Attach_User_Context(
    struct Kernel_Thread* kthread, struct User_Context* context )
{
    // FIXME: race condition

    kthread->userContext = context;
    if ( context != 0 ) {
	++context->refCount;
    }
}

// Detach a user context from the given thread.
// Called with interrupts enabled.
void Detach_User_Context( struct Kernel_Thread* kthread )
{
    // FIXME: race condition

    struct User_Context* old = kthread->userContext;

    kthread->userContext = 0;

    if ( old != 0 ) {
	--old->refCount;
	if ( old->refCount == 0 )
	    Destroy_User_Context( old );
    }
}

// Switch user context (if needed).  This is called when a new thread
// is scheduled, before returning to it.  Interrupts must be disabled.
void Switch_User_Context(
	struct Kernel_Thread* kthread, struct Interrupt_State* state )
{
    struct User_Context* userContext = kthread->userContext;
    int rpl;

    KASSERT( !Interrupts_Enabled() );

    // We only need to activate the user context if we are
    // in fact returning to user mode.  The RPL of the code segment
    // selector will tell us what privilege level we're returning to.
    rpl = (state->cs & 0x3);
 
    if ( rpl == KERNEL_PRIVILEGE ) {
	//Print( "[RK %x]\n", kthread );
	return; // returning to kernel mode
    }

    KASSERT( rpl == USER_PRIVILEGE );
    KASSERT( userContext != 0 );
    KASSERT( (state->eflags & EFLAGS_IF) == EFLAGS_IF );

    // Switch only if the user context is indeed different
    if ( userContext != s_currentUserContext ) {
	unsigned short ldtSelector;
	unsigned long esp0;

	//Print( "A[%x]\n", kthread );

	// Switch to the LDT of the new user context
	ldtSelector = userContext->ldtSelector;
	__asm__ __volatile__ (
	    "lldt %0"
	    :
	    : "a" (ldtSelector)
	);


	// By definition, when returning to user mode there is no
	// context remaining on the kernel stack.
	esp0 = ((unsigned long) kthread->stackPage) + PAGE_SIZE;
	//Print( "S[%x]\n", esp0 );

	// The processor will use the TSS to determing which ring-0
	// stack to use if an interrupt occurs in user mode
	g_theTSS.esp0 = esp0;
	g_theTSS.ss0 = KERNEL_DS;

	// New user context is active
	s_currentUserContext = userContext;
    }
}

#define DEFAULT_USER_STACK_SIZE 4096

// Start a user-mode thread from a User_Program object.
struct Kernel_Thread* Start_User_Program( const struct User_Program* program,
	Boolean detached )
{
    struct User_Context* userContext;

    // Create the user context, with some room for the program's stack.
    userContext = Create_User_Context( program->size + DEFAULT_USER_STACK_SIZE );
    if ( userContext == 0 )
	return FALSE;

    // Initialize the user program memory to zero, and
    // copy the program code and data into it.
    memset( userContext->memory, '\0', userContext->size );
    memcpy( userContext->memory, program->data, program->size );

    // Fire it up!
    return Start_User_Thread( userContext, program->entryAddr, TRUE );
}

// Validate that buffer at given user-space address within given
// user context lies entirely within the user context's memory.
// Used to ensure that Copy_From_User() and Copy_To_User() do not
// access memory outside the user context.
Boolean Validate_User_Memory( struct User_Context* userContext,
    void* userPtr, unsigned long bufSize )
{
    unsigned long userAddr = (unsigned long) userPtr;
    unsigned long avail;

    if ( userAddr >= userContext->size )
	return FALSE;

    avail = userContext->size - userAddr;
    if ( bufSize > avail )
	return FALSE;

    return TRUE;
}

// Copy data from current user context into given kernel
// buffer.  Returns true if successful, or false if any part
// of the source buffer lies outside the user context memory.
Boolean Copy_From_User( void* destInKernel, const void* srcInUser,
	unsigned long bufSize )
{
    struct User_Context* current = g_currentThread->userContext;

    if ( !Validate_User_Memory( current, (void*) srcInUser, bufSize ) )
	return FALSE;
    memcpy( destInKernel, User_To_Kernel(current, (void*)srcInUser), bufSize );

    return TRUE;
}

// Copy data from kernel buffer into current user context at
// given address.  Returns true if successful, or false if
// any part of the destination buffer lies outside the
// use context memory.
Boolean Copy_To_User( void* destInUser, const void* srcInKernel,
	unsigned long bufSize )
{
    struct User_Context* current = g_currentThread->userContext;

    if ( !Validate_User_Memory( current, destInUser, bufSize ) )
	return FALSE;
    memcpy( User_To_Kernel(current, destInUser), srcInKernel, bufSize );

    return TRUE;
}
